from django.test import TestCase, Client
from django.core.exceptions import ValidationError
from ..models import Sale, Product, SaleLine, Customer, ReconciliationJar

import json


class TestJarReconcliation(TestCase):
    def setUp(self):
        customer = Customer()
        customer.name = 'Alejo Mono'
        customer.save()

        self.client = Client()

        purchase = Sale()
        purchase.customer = customer
        purchase.date = "2024-07-30"
        purchase.payment_method = 'CASH'
        purchase.clean()
        purchase.save()

        product = Product()
        product.name = "cafe"
        product.price = "72500"
        product.save()

        line = SaleLine()
        line.sale = purchase
        line.product = product
        line.quantity = "11"
        line.unit_price = "72500"
        line.save()
        self.purchase = purchase

        purchase2 = Sale()
        purchase2.customer = customer
        purchase2.date = "2024-07-30"
        purchase.payment_method = 'CASH'
        purchase2.clean()
        purchase2.save()

        line2 = SaleLine()
        line2.sale = purchase2
        line2.product = product
        line2.quantity = "27"
        line2.unit_price = "72500"
        line2.save()
        self.purchase2 = purchase2

        purchase3 = Sale()
        purchase3.customer = customer
        purchase3.date = "2024-07-30"
        purchase3.payment_method = 'CASH'
        purchase3.clean()
        purchase3.save()

        line3 = SaleLine()
        line3.sale = purchase3
        line3.product = product
        line3.quantity = "37"
        line3.unit_price = "72500"
        line3.save()
        self.purchase3 = purchase3

        purchase4 = Sale()
        purchase4.customer = customer
        purchase4.date = "2024-07-30"
        purchase4.payment_method = 'CONFIAR'
        purchase4.clean()
        purchase4.save()

        line4 = SaleLine()
        line4.sale = purchase4
        line4.product = product
        line4.quantity = "47"
        line4.unit_price = "72500"
        line4.save()
        self.purchase4 = purchase4

    def test_create_reconciliation_jar(self):
        reconciliation = self._create_simple_reconciliation()
        self.assertTrue(isinstance(reconciliation, ReconciliationJar))

    def test_get_purchases_for_reconciliation(self):
        # link purchase to reconciliation to exclude from list
        reconciliation = self._create_simple_reconciliation()
        self.purchase3.reconciliation = reconciliation
        self.purchase3.clean()
        self.purchase3.save()

        url = '/don_confiao/purchases/for_reconciliation'
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        rawContent = response.content.decode('utf-8')
        content = json.loads(rawContent)

        self.assertIn('CASH', content.keys())
        self.assertIn('CONFIAR', content.keys())
        self.assertEqual(2, len(content.get('CASH')))
        self.assertEqual(1, len(content.get('CONFIAR')))
        self.assertNotIn(str(37*72500), rawContent)
        self.assertIn(str(47*72500), rawContent)

    def test_don_create_reconcialiation_with_bad_numbers(self):
        reconciliation = ReconciliationJar()
        reconciliation.date_time = "2024-07-30"
        reconciliation.total_cash_purchases = 145000
        reconciliation.cash_taken = 143000
        reconciliation.cash_discrepancy = 1000
        with self.assertRaises(ValidationError):
            reconciliation.clean()
            reconciliation.save()

    def test_fail_create_reconciliation_with_wrong_total_purchases_purchases(self):
        url = '/don_confiao/reconciliate_jar'
        total_purchases = (11 * 72500) + (27 * 72500)
        bad_total_purchases = total_purchases + 2
        data = {
            'date_time': '2024-12-02T21:07',
            'reconcilier': 'carlos',
            'total_cash_purchases': bad_total_purchases,
            'cash_taken': total_purchases,
            'cash_discrepancy': 0,
            'cash_purchases': [
                self.purchase.id,
                self.purchase2.id,
                self.purchase.id,
            ],
        }
        response = self.client.post(url, data=json.dumps(data).encode('utf-8'),
                                    content_type='application/json')
        rawContent = response.content.decode('utf-8')
        content = json.loads(rawContent)

        self.assertEqual(response.status_code, 400)
        self.assertIn('error', content)
        self.assertIn('total_cash_purchases', content['error'])

    def test_create_reconciliation_with_purchases(self):
        response = self._create_reconciliation_with_purchase()

        rawContent = response.content.decode('utf-8')
        content = json.loads(rawContent)

        self.assertEqual(response.status_code, 200)
        self.assertIn('id', content)

        purchases = Sale.objects.filter(reconciliation_id=content['id'])
        self.assertEqual(len(purchases), 2)

    def test_create_reconciliation_with_purchases_and_other_totals(self):
        url = '/don_confiao/reconciliate_jar'
        total_purchases = (11 * 72500) + (27 * 72500)
        data = {
            'date_time': '2024-12-02T21:07',
            'reconcilier': 'carlos',
            'total_cash_purchases': total_purchases,
            'cash_taken': total_purchases,
            'cash_discrepancy': 0,
            'cash_purchases': [
                self.purchase.id,
                self.purchase2.id,
            ],
            'other_totals': {
                'Confiar': {
                    'total': (47 * 72500) + 1,
                    'purchases': [self.purchase4.id],
                },
            },
        }
        response = self.client.post(url, data=json.dumps(data).encode('utf-8'),
                                    content_type='application/json')

        rawContent = response.content.decode('utf-8')
        content = json.loads(rawContent)

        self.assertEqual(response.status_code, 200)
        self.assertIn('id', content)

        purchases = Sale.objects.filter(reconciliation_id=content['id'])
        self.assertEqual(len(purchases), 3)

    def test_list_reconciliations(self):
        self._create_simple_reconciliation()
        self._create_simple_reconciliation()

        url = '/don_confiao/api/reconciliate_jar/'

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        content = json.loads(response.content.decode('utf-8'))
        self.assertEqual(2, content['count'])
        self.assertEqual(2, len(content['results']))
        self.assertEqual('2024-07-30T00:00:00Z',
                         content['results'][0]['date_time'])

    def test_list_reconciliations_pagination(self):
        self._create_simple_reconciliation()
        self._create_simple_reconciliation()

        url = '/don_confiao/api/reconciliate_jar/?page=2&page_size=1'

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        content = json.loads(response.content.decode('utf-8'))
        self.assertEqual(1, len(content['results']))
        self.assertEqual('2024-07-30T00:00:00Z',
                         content['results'][0]['date_time'])

    def test_get_single_reconciliation(self):
        createResponse = self._create_reconciliation_with_purchase()
        reconciliationId = json.loads(
            createResponse.content.decode('utf-8')
        )['id']
        self.assertGreater(reconciliationId, 0)

        url = f'/don_confiao/api/reconciliate_jar/{reconciliationId}/'
        response = self.client.get(url, content_type='application/json')
        content = json.loads(
            response.content.decode('utf-8')
        )
        self.assertEqual(reconciliationId, content['id'])
        self.assertGreater(len(content['Sales']), 0)
        self.assertIn(
            self.purchase.id,
            [sale['id'] for sale in content['Sales']]
        )

        self.assertIn(
            'CASH',
            [sale['payment_method'] for sale in content['Sales']]
        )

    def _create_simple_reconciliation(self):
        reconciliation = ReconciliationJar()
        reconciliation.date_time = "2024-07-30"
        reconciliation.total_cash_purchases = 0
        reconciliation.cash_taken = 0
        reconciliation.cash_discrepancy = 0
        reconciliation.clean()
        reconciliation.save()
        return reconciliation

    def _create_reconciliation_with_purchase(self):
        url = '/don_confiao/reconciliate_jar'
        total_purchases = (11 * 72500) + (27 * 72500)
        data = {
            'date_time': '2024-12-02T21:07',
            'reconcilier': 'carlos',
            'total_cash_purchases': total_purchases,
            'cash_taken': total_purchases,
            'cash_discrepancy': 0,
            'cash_purchases': [
                self.purchase.id,
                self.purchase2.id,
                self.purchase.id,
            ],
        }
        return self.client.post(url, data=json.dumps(data).encode('utf-8'),
                                content_type='application/json')
